/*
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Copyright 2000 by Michael B. Allen
 *
 */

package jcifs.netbios;

import java.net.*;

class ClientNameServicePacket {

    static final int MAX_DATAGRAM_LENGTH = 576;
    static final int NAME_SERVICE_PORT = 137;

    // packet types
    static final int POS_NAME_QUERY_RESPONSE = 0;
    static final int NEG_NAME_QUERY_RESPONSE = 1;
    static final int REDIRECT_NAME_QUERY_RESPONSE = 2;
    static final int WAIT_FOR_ACK_RESPONSE = 3;
    static final int NODE_STATUS_RESPONSE = 4;
    static final int NAME_QUERY_REQUEST = 5;
    static final int NODE_STATUS_REQUEST = 6;
    static final int OTHER = 7;

    /*
     * Name Query Request
     *
     * RO   ATR R  BR
     *  C   A D A   C
     * 00000001 00010000
     */

    static final byte[] HDR_NAME_QUERY_REQUEST = {
                                (byte)0x01, (byte)0x10,
        (byte)0x00, (byte)0x01, (byte)0x00, (byte)0x00,
        (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00
    };

    /*
     * Node Status Request
     *
     * RO   ATR R  BR
     *  C   A D A   C
     * 00000000 00000000
     */

    static final byte[] HDR_NODE_STATUS_REQUEST = {
                                (byte)0x00, (byte)0x00,
        (byte)0x00, (byte)0x01, (byte)0x00, (byte)0x00,
        (byte)0x00, (byte)0x00, (byte)0x00, (byte)0x00,
        (byte)0x20, (byte)0x43, (byte)0x4B, (byte)0x41,
        (byte)0x41, (byte)0x41, (byte)0x41, (byte)0x41,
        (byte)0x41, (byte)0x41, (byte)0x41, (byte)0x41,
        (byte)0x41, (byte)0x41, (byte)0x41, (byte)0x41,
        (byte)0x41, (byte)0x41, (byte)0x41, (byte)0x41,
        (byte)0x41, (byte)0x41, (byte)0x41, (byte)0x41,
        (byte)0x41, (byte)0x41, (byte)0x41, (byte)0x41,
        (byte)0x41, (byte)0x41, (byte)0x41, (byte)0x41,
        (byte)0x41, (byte)0x00
    };

    static final int    NBSTAT    = 0x0021;
    static final byte[] NB_IN     = { (byte)0x00, (byte)0x20, (byte)0x00, (byte)0x01 };
    static final byte[] NBSTAT_IN = { (byte)0x00, (byte)0x21, (byte)0x00, (byte)0x01 };
    static final int    QUERY     = 0;
    static final int    WACK      = 7;

    Name name = null;
    DatagramPacket packet;
    byte[] bufr;
    static int nameTrnId = 0;

    ClientNameServicePacket() {
        bufr = new byte[MAX_DATAGRAM_LENGTH];
        packet = new DatagramPacket( bufr, MAX_DATAGRAM_LENGTH );
        packet.setPort( NAME_SERVICE_PORT );
    }
    ClientNameServicePacket( byte[] bufr, int length ) {
        this.bufr = bufr;
        packet = new DatagramPacket( bufr, length );
        packet.setPort( NAME_SERVICE_PORT );
    }
    ClientNameServicePacket( byte[] bufr, int length, InetAddress addr, int port ) {
        this.bufr = bufr;
        packet = new DatagramPacket( bufr, length, addr, port );
        packet.setPort( NAME_SERVICE_PORT );
    }
    ClientNameServicePacket( DatagramPacket p ) {
        packet = p;
        bufr = packet.getData();
        packet.setPort( NAME_SERVICE_PORT );
    }

    // DatagramPacket methods

    synchronized InetAddress getAddress() {
        return packet.getAddress();
    }
    synchronized int getPort() {
        return packet.getPort();
    }
    synchronized byte[] getData() {
        return packet.getData();
    }
    synchronized int getLength() {
        return packet.getLength();
    }
    synchronized void setAddress( InetAddress addr ) {
        packet.setAddress( addr );
    }
    synchronized void setPort( int port ) {
        packet.setPort( port );
    }
    synchronized void setData( byte[] bufr ) {
        packet.setData( bufr );
    }
    synchronized void setLength( int length ) {
        packet.setLength( length );
    }

    int writeNameQueryRequest() {
        setNameTrnId();
        System.arraycopy( HDR_NAME_QUERY_REQUEST, 0, bufr, 2, 10 );
        int index = 12;
        index += name.writeFullName( bufr, index );
        System.arraycopy( NB_IN, 0, bufr, index, 4 );
        index += 4;
        setLength( index );
        return index;
    }
    int writeNodeStatusRequest() {
        setNameTrnId();
        System.arraycopy( HDR_NODE_STATUS_REQUEST, 0, bufr, 2, 43 );
        int index = 45;
        index += name.writeScope( bufr, 45 );
        System.arraycopy( NBSTAT_IN, 0, bufr, index, 4 );
        index += 4;
        setLength( index );
        return index;
    }

    // header get and is methods

    int getNameTrnId() {
        return (( bufr[0] & 0xFF ) << 8 ) + ( bufr[1] & 0xFF );
    }
    boolean isResponse() {
        return (( bufr[2] & 0x80 ) == 0 ) ? false : true;
    }
    int getOpCode() {
        return ( bufr[2] & 0x78 ) >> 3;
    }
    boolean isAuthAnswer() {
        return (( bufr[2] & 0x04 ) == 0 ) ? false : true;
    }
    boolean isTruncated() {
        return (( bufr[2] & 0x02 ) == 0 ) ? false : true;
    }
    boolean isRecurDesired() {
        return (( bufr[2] & 0x01 ) == 0 ) ? false : true;
    }
    boolean isRecurAvailable() {
        return (( bufr[3] & 0x80 ) == 0 ) ? false : true;
    }
    boolean isBroadcast() {
        return (( bufr[3] & 0x10 ) == 0 ) ? false : true;
    }
    int getResultCode() {
        return bufr[3] & 0x0F;
    }
    int getQuestionCount() {
        return (( bufr[4] & 0xFF ) << 8 ) + ( bufr[5] & 0xFF );
    }
    int getAnswerCount() {
        return (( bufr[6] & 0xFF ) << 8 ) + ( bufr[7] & 0xFF );
    }
    int getAuthorityCount() {
        return (( bufr[8] & 0xFF ) << 8 ) + ( bufr[9] & 0xFF );
    }
    int getAdditionalCount() {
        return (( bufr[10] & 0xFF ) << 8 ) + ( bufr[11] & 0xFF );
    }
    // these 2 are really only for pos name query for Log.printPacketDiagram()
    boolean isGroupName() {
        int pos = Name.wireLength( bufr, 12 ) + 22;
        return (( bufr[pos] & 0x80 ) == 0x80 ) ? true : false;
    }
    int getOwnerNodeType() {
        int pos = Name.wireLength( bufr, 12 ) + 22;
        return ( bufr[pos] & 0x60 ) >> 5;
    }

    // body get and is methods

    int getRecordType() {
        // returns only the first record type
        // but thats all this would really be used for
        // which is to tighten the validation during
        // getPacketType for packets that have uniq
        // first record types

        int pos = 12 + Name.wireLength( bufr, 12 );
        return (( bufr[pos] << 8 ) & 0xFF ) + ( bufr[pos + 1] & 0xFF );
    }

    // really just for WACK

    int getTTL() {
        int pos = Name.wireLength( bufr, 12 ) + 16;
        int ttl = ( bufr[pos] << 24 ) & 0xFF000000;
        ttl |= ( bufr[pos + 1] << 16 ) & 0xFF0000;
        ttl |= ( bufr[pos + 2] << 8 ) & 0xFF00;
        ttl |= bufr[pos + 3] & 0xFF;
        return ttl;
    }
    NbtAddress getNbtAddress( Name name ) {
        int pos = packet.getLength();
        int address = ( bufr[pos - 4] << 24 ) & 0xFF000000;
        address |= ( bufr[pos - 3] << 16 ) & 0xFF0000;
        address |= ( bufr[pos - 2] << 8 ) & 0xFF00;
        address |= bufr[pos - 1] & 0xFF;
        boolean groupName = (( bufr[pos - 6] & 0x80 ) == 0x80 ) ? true : false;
        int nodeType = ( bufr[pos - 6] & 0x60 ) >> 5;
        return new NbtAddress( name, address, groupName, nodeType );
    }
    InetAddress getInetAddress() {
        int pos = packet.getLength();
        StringBuffer sb = new StringBuffer( 15 );
        sb.append(( bufr[pos - 4] & 0xFF ) + "." );
        sb.append(( bufr[pos - 3] & 0xFF ) + "." );
        sb.append(( bufr[pos - 2] & 0xFF ) + "." );
        sb.append(( bufr[pos - 1] & 0xFF ) + "" );
        try {
            return InetAddress.getByName( sb.toString() );
        } catch( UnknownHostException uhe ) {
            // should never make it here
        }
        return null;
    }

    NbtAddress[] getNodeNameArray( NbtAddress addr ) {

        Name hostName = addr.hostName;

        // find starting pos of node name array

        int pos = Name.wireLength( bufr, 12 ) + 22;

        // get number of names in array and allocate array to return

        int numNames = bufr[pos++] & 0xFF;
        NbtAddress addrs[] = new NbtAddress[numNames];

        // get the mac address

        byte mac[] = new byte[6];
        int mPos = pos + numNames * 18;
        System.arraycopy( bufr, mPos, mac, 0, 6 );
        addr.macAddress = mac;

        // find all the stuff from each entry and use it to create
        // a new NbtAddress for addition into the array addrs

        String n;
        int type;
        String scope = ( hostName.scope != null ) ? hostName.scope : null;
        boolean groupName;
        int ownerNodeType;
        boolean isBeingDeleted;
        boolean isInConflict;
        boolean isActive;
        boolean isPermanent;
        int j;
        boolean addrFound = false;
        for( int i = 0; i < numNames; pos += 18, i++ ) {
            for( j = pos + 14; bufr[j] == 0x20; j-- )
                ;
            n = new String( bufr, pos, j - pos + 1 );
            type            =    bufr[pos + 15] & 0xFF;
            groupName        = (( bufr[pos + 16] & 0x80 ) == 0x80 ) ? true : false;
            ownerNodeType    =  ( bufr[pos + 16] & 0x60 ) >> 5;
            isBeingDeleted    = (( bufr[pos + 16] & 0x10 ) == 0x10 ) ? true : false;
            isInConflict    = (( bufr[pos + 16] & 0x08 ) == 0x08 ) ? true : false;
            isActive        = (( bufr[pos + 16] & 0x04 ) == 0x04 ) ? true : false;
            isPermanent        = (( bufr[pos + 16] & 0x02 ) == 0x02 ) ? true : false;
            if( !addrFound && hostName.type == type &&
                    ( hostName.name == null || hostName.name.equalsIgnoreCase( n ))) {
                addr.hostName.setName( n );
                addr.groupName = groupName;
                addr.nodeType = ownerNodeType;
                addr.isBeingDeleted = isBeingDeleted;
                addr.isInConflict = isInConflict;
                addr.isActive = isActive;
                addr.isPermanent = isPermanent;
                addr.macAddress = mac;
                addr.isDataFromNodeStatus = true;
                addrFound = true;
                addrs[i] = addr;
            } else {
                addrs[i] = new NbtAddress( new Name( n, type, scope ),
                                        addr.address,
                                        groupName,
                                        ownerNodeType,
                                        isBeingDeleted,
                                        isInConflict,
                                        isActive,
                                        isPermanent,
                                        mac );
            }
        }
        return addrs;
    }

    // header set methods operate on bufr directly
    // only really need a few of these

    void setNameTrnId() {
        bufr[0] = (byte)(( nameTrnId & 0xFF00 ) >> 8 );
        bufr[1] = (byte)( nameTrnId & 0xFF );
    }
    void setName( Name name ) {
        this.name = name;
    }
    void setRecurDesired( boolean recurDesired ) {
        if( recurDesired ) {
            bufr[2] &= 0xFE;
            bufr[2] += 0x01;
        } else {
            bufr[2] &= 0xFE;
        }
    }
    void setBroadcast( boolean broadcast ) {
        if( broadcast ) {
            bufr[3] &= 0xEF;
            bufr[3] += 0x10;
        } else {
            bufr[3] &= 0xEF;
        }
    }

    // determine a packets type from wire format

    int getPacketType() {
        if( isResponse() ) {
            switch( getOpCode() ) {
                case QUERY:
                    if( getResultCode() == 0x0 ) {
                        if( !isRecurDesired() ) {
                            return NODE_STATUS_RESPONSE;
                        } else if( isAuthAnswer() ) {
                            return POS_NAME_QUERY_RESPONSE;
                        }
                        return REDIRECT_NAME_QUERY_RESPONSE;
                    }
                    return NEG_NAME_QUERY_RESPONSE;
                case WACK:
                    return WAIT_FOR_ACK_RESPONSE;
            }
        } else {
            switch( getOpCode() ) {
                case QUERY:
                    if( getRecordType() == NBSTAT ) {
                        return NODE_STATUS_REQUEST;
                    }
                    return NAME_QUERY_REQUEST;
            }
        }
        return OTHER;
    }
}

