/*
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Copyright 2000 by Michael B. Allen
 *
 */

package jcifs.netbios;

import java.net.*;
import java.io.IOException;
import java.io.PrintWriter;
import jcifs.util.MultiDatagramSocket;
import jcifs.util.PacketEventQueue;
import jcifs.util.PacketEvent;

/**
 * Applications should use the {@link jcifs.netbios.NbtAddress} class for
 * interacting with the netbios. This class is for lower-level control over
 * the name service and would only be usefull for netbios centric tools.
 * <p> This class provides the name service functionality for the netbios
 * networking protocol. The netbios name service is a dynamic distributed
 * service that allows hosts to resolve names by broadcasting a query,
 * directing querys to a server such as Samba or WINS. A netbios name
 * consists of two parts, a 15 character string and a 1 byte type. A netbios
 * node usually registers several names on the network. You can see what
 * names a Windows machine registers with the nbtstat command.
 * <p><blockquote><pre>
 * C:\>nbtstat -a 192.168.1.15
 * 
 *        NetBIOS Remote Machine Name Table
 * 
 *    Name               Type         Status
 * ---------------------------------------------
 * ANGUS          <00>  UNIQUE      Registered
 * RSCH-NY        <00>  GROUP       Registered
 * ANGUS          <03>  UNIQUE      Registered
 * ANGUS          <20>  UNIQUE      Registered
 * RSCH-NY        <1E>  GROUP       Registered
 * MIALLEN        <03>  UNIQUE      Registered
 * 
 * MAC Address = 00-A0-34-21-FA-3B
 * </blockquote></pre>
 * <p> The hostname of this machine is <code>ANGUS</code>. It is
 * a member of the group(or workgroup) <code>RSCH-NY</code>. The
 * hexadecimal values are the names types. The types correspond to
 * the services that the host provides. Usually the <code><00></code>
 * type is the only of interest as it is registered by all hosts. <a
 * href="http://users.erols.com/mballen/NetBiosHexCodes.txt">Here</a>
 * is a list of what services correspond to what hex codes.
 *
 * @version   0.1, 00/03/13
 * @author    Michael B. Allen
 * @since     jcifs-0.1
 */

public class NbtNameServiceClient {

    // TIMING
    static final int BCAST_REQ_RETRY_COUNT       = 3;
    static final int UCAST_REQ_RETRY_COUNT       = 3;
    static final int BCAST_REQ_RETRY_TIMEOUT     = 500;
    static final int UCAST_REQ_RETRY_TIMEOUT     = 5000;
    static final int BCAST_REQ_DURATION          = 500;
    static final int MAX_UCAST_REQ_RETRY_TIMEOUT = 15000;
    static final int MAX_REDIRECT_COUNT          = 10;

    // TRANSMISSION TYPES
    static final int BROADCAST = 0;
    static final int UNICAST   = 1;

    private int maxRedirectCount;

    static MultiDatagramSocket socket;
    InetAddress nameserver = null;
    InetAddress bCastAddress = null;
    NbtAddress localhost = null;
    int nodeType = -1;

/**
 * Instantiate a new client. The name service client will not bind the port
 * until a request is issued. The underlying socket will timeout and close
 * after 30 seconds of incativity.
 *
 * @throws java.net.SocketException if the {@link jcifs.util.MultiDatagramSocket} cannot be started
 */

    public NbtNameServiceClient() throws SocketException {
        this( -1, null );
    }

/**
 * Instantiate a new client. The name service client will not bind the port
 * until a request is issued. The underlying socket will timeout and close
 * after 30 seconds of incativity.
 *
 * @param port the port the client should bind to.
 * @throws java.net.SocketException if the {@link jcifs.util.MultiDatagramSocket} cannot be started
 */

    public NbtNameServiceClient( int port ) throws SocketException {
        this( port, null );
    }

/**
 * Instantiate a new client. The name service client will not bind the port
 * until a request is issued. The underlying socket will timeout and close
 * after 30 seconds of incativity.
 * 
 * @param port the port the client should run on
 * @param laddr the address of the interface the client should bind
 * @throws java.net.SocketException if the {@link jcifs.util.MultiDatagramSocket} cannot be started
 */

    public NbtNameServiceClient( int port, InetAddress laddr ) throws SocketException {
        if( port < 1 && laddr == null ) {
            socket = new MultiDatagramSocket();
        } else if( laddr == null ) {
            socket = new MultiDatagramSocket( port );
        } else {
            socket = new MultiDatagramSocket( port, laddr );
        }
    }

    private void guessBroadcastAddress() throws UnknownNbtHostException {
        try {
            byte bcast[] = InetAddress.getLocalHost().getAddress();
            bcast[3] = (byte)255;
            byte tmp[] = new byte[15];
            int i = 14;
            int val = 0;

            for( int j = 3; j >= 0; j-- ) {
                val = bcast[j] & 0xFF;
                do {
                    tmp[i--] = (byte)(( val % 10 ) + 48 );
                    val = val / 10;
                } while( val > 0 );
                tmp[i--] = '.';
            }
            i += 2;
            bCastAddress = InetAddress.getByName( new String( tmp, i, 15 - i ));
        } catch( UnknownHostException uhe ) {
            bCastAddress = null;
            throw new UnknownNbtHostException( "broadcast address is null" );
        }
    }
    private void guessNodeType() {
        if( nameserver != null ) {
            nodeType = NbtAddress.H_NODE;
        } else {
            nodeType = NbtAddress.B_NODE;
        }
    }

/**
 * Set the local host name that the client uses when connecting to other
 * hosts. The scope is likely not used and can therefore be and empty string
 * or null. If the <code>name</code> parameter is empty or null a name will
 * try resolve the name of the machine using standard netbios queries. If
 * that fails the first label of the DNS name is use. If that fails the name
 * <code>JCIFS_XXX</code> is used where <code>XXX</code> is a random number.
 * 
 * @param name the name this client should masquarade as
 * @param scope the scope of the client's name. This is generally not used but if it is, it must be set
 */

    public void setLocalHostName( String name, String scope ) {

        // start with the localhost address of course
        InetAddress inet = null;
        NbtAddress addr = null;
        try {
            inet = InetAddress.getLocalHost();
            // first get NbtAddress to work on using local host address string
            addr = NbtAddress.getByName( inet.getHostAddress(), 0x00, scope );
        } catch( UnknownHostException uhe ) {
            Log.printStackTrace( "jcifs.netbios.NbtNameServiceClient.setLocalHostName()", uhe );
        }
        // use specified name first
        if( name != null && name.length() != 0 ) {
            // just set the name explicitly in the addr and we're done
            addr.hostName.setName( name );
            localhost = addr;
            NbtAddress.cacheAddress( localhost, NbtAddress.FOREVER );
            return;
        }
        // if name not specified let's try to resolve using local system netbios
        String h = addr.getHostName();
        if( !Character.isDigit( h.charAt( 0 ))) {
            // got it, we're done
            localhost = addr;
            NbtAddress.cacheAddress( localhost );
            return;
        }
        // local machine doesn't have netbios end node listening so lets try the dns name
        String n = inet.getHostName();
        // does the dns name start with chars
        if( !Character.isDigit( n.charAt( 0 ))) {
            // if so, get the first label
            int j;
            if(( j = n.indexOf( '.' )) != -1 ) {
                j = ( j > 11 ) ? 11 : j;
                n = n.substring( 0, j );
            } else {
                n = n.length() > 11 ? n.substring( 0, 11 ) : n;
            }
        } else {
            // if the dns name doesn't start with a char lets just make up a name
            n = new String( "JCIFS" );
        }
        // now append a number from 0 - 999 to make it uniqe
        int time = (int)System.currentTimeMillis();
        n += new String( "_" + ( time & 0x3e7 ));
        addr.hostName.setName( n );
        localhost = addr;
        NbtAddress.cacheAddress( localhost );
    }

/**
 * Retrieve the local host name.
 * @see jcifs.netbios.NbtNameServiceClient#setLocalHostName(String name, String scope)
 */

    public NbtAddress getLocalHostName() {
        if( localhost == null ) {
            setLocalHostName( null, null );
        }
        return localhost;
    }

/** 
 * Determines the address of a host given it's host name. The name can be a
 * machine name like "jcifs.samba.org" or an IP address like "192.168.1.15".
 *
 * @param name hostname to resolve
 * @throws jcifs.netbios.UnknownNbtHostException if there is an error resolving the name
 */

    public NbtAddress getByName( String name )
                                throws UnknownNbtHostException {

        return getByName( new Name( name, 0, "" ), true );
    }

/** 
 * Determines the address of a host given it's host name. Netbios
 * names also have a <code>type</code>. Types(aka Hex Codes)
 * are used to distiquish the various services on a host. <a
 * href="http://users.erols.com/mballen/NetBiosHexCodes.txt">Here</a> is
 * a fairly complete list of netbios hex codes. Scope is not used but is
 * still functional on Windows machines and so for completeness it has been
 * implemented. A <code>scope</code> of <code>null</code> or <code>""</code>
 * signifies no scope.
 *
 * @param name the name to resolve
 * @param type the hex code of the name
 * @param scope the scope of the name
 * @param is recursion desired
 * @throws jcifs.netbios.UnknownNbtHostException if there is an error resolving the name
 */

    public NbtAddress getByName( String name,
                                int type,
                                String scope,
                                boolean recurDesired )
                                throws UnknownNbtHostException {

        return getByName( new Name( name, type, scope ), recurDesired );
    }

    NbtAddress getByName( Name name, boolean recurDesired )
                                throws UnknownNbtHostException {

        ClientNameServicePacket.nameTrnId++;
        ClientNameServicePacket nsp = new ClientNameServicePacket();
        nsp.setName( name );
        nsp.writeNameQueryRequest();
        nsp.setRecurDesired( recurDesired );
        NbtAddress addr = null;
        
        if( nodeType == -1 ) {
            guessNodeType();
        }
        try {
            switch( nodeType ) {
                case NbtAddress.H_NODE:
                    if( nameserver == null ) {
                        throw new UnknownNbtHostException( "name server address is null" );
                    }
                    if( bCastAddress == null ) {
                        guessBroadcastAddress();
                    }
                    addr = doUnicast( nsp );
                    if( addr != null ) {
                        return addr;
                    }
                    addr = doBroadcast( nsp );
                    if( addr != null ) {
                        return addr;
                    }
                    break;
                case NbtAddress.P_NODE:
                    if( nameserver == null ) {
                        throw new UnknownNbtHostException( "name server address is null" );
                    }
                    addr = doUnicast( nsp );
                    if( addr != null ) {
                        return addr;
                    }
                    break;
                case NbtAddress.M_NODE:
                    if( nameserver == null ) {
                        throw new UnknownNbtHostException( "name server address is null" );
                    }
                    if( bCastAddress == null ) {
                        guessBroadcastAddress();
                    }
                    addr = doBroadcast( nsp );
                    if( addr != null ) {
                        return addr;
                    }
                    addr = doUnicast( nsp );
                    if( addr != null ) {
                        return addr;
                    }
                    break;
                default:
                    if( bCastAddress == null ) {
                        guessBroadcastAddress();
                    }
                    addr = doBroadcast( nsp );
                    if( addr != null ) {
                        return addr;
                    }
            }
        } catch( Exception e ) {
            Log.printStackTrace( "jcifs.netbios.NbtNameServiceClient.getByName()", e );
        }
        throw new UnknownNbtHostException( name.toString() );
    }

/** 
 * Resloves the addresses of all hosts given a name(presumably a workgroup
 * name).
 *
 * @param name groupname to lookup
 * @throws jcifs.netbios.UnknownNbtHostException if there is an error resolving the name
 */

    public NbtAddress[] getAllByName( String name )
                                    throws UnknownNbtHostException {
        return getAllByName( new Name( name, 0, "" ), true );
    }

/** 
 * Resloves the addresses of all hosts given a hostname(presumably a
 * workgroup name).  See {@link #getByName}
 * for a description of <code>type</code> and <code>scope</code>.
 *
 * @param name group name to lookup
 * @param type the hexcode of the name
 * @param is recursion desired
 * @throws jcifs.netbios.UnknownNbtHostException if there is an error resolving the name
 */

    public NbtAddress[] getAllByName( String name,
                                    int type,
                                    String scope,
                                    boolean recurDesired )
                                    throws UnknownNbtHostException {

        return getAllByName( new Name( name, type, scope ),
                                recurDesired );
    }

    NbtAddress[] getAllByName( Name name, boolean recurDesired )
                                    throws UnknownNbtHostException {

        if( bCastAddress == null ) {
            guessBroadcastAddress();
        }
        ClientNameServicePacket.nameTrnId++;
        ClientNameServicePacket nsp = new ClientNameServicePacket();
        nsp.setName( name );
        nsp.writeNameQueryRequest();
        nsp.setRecurDesired( recurDesired );
        nsp.setAddress( bCastAddress );

        NbtAddress[] addrs = null;

        try {
            addrs = sendAndReceiveAll( nsp );
        } catch( IOException ioe ) {
            addrs = null;
        } catch( InterruptedException ie ) {
            socket = null;
            Log.printStackTrace( "jcifs.netbios.NbtNameServiceClient.getAllByName()", ie );
        }
        if( addrs != null ) {
            return addrs;
        }
        throw new UnknownNbtHostException( name.toString() );
    }

/**
 * Do a node status on an address.
 *
 * @param addr the address to do a node status on
 * @see jcifs.netbios.NbtAddress#getAllByAddress( NbtAddress addr )
 * @throws jcifs.netbios.UnknownNbtHostException if the address does not respond
 */

    public NbtAddress[] getNodeStatus( NbtAddress addr )
                                        throws UnknownNbtHostException {
        ClientNameServicePacket out = new ClientNameServicePacket();
        out.name = addr.hostName;
        out.writeNodeStatusRequest();
        try {
            out.setAddress( addr.getInetAddress() );
        } catch( UnknownHostException uhe ) {
            throw new UnknownNbtHostException();
        }
        int retryCount = UCAST_REQ_RETRY_COUNT;
        int retryTimeout;
        try {
            socket.start();
        } catch( SocketException se ) {
            Log.printStackTrace( "jcifs.netbios.NbtNameServiceClient.getNodeStatus()", se );
            throw new UnknownNbtHostException( "socket error" );
        }
        PacketEventQueue peq = socket.createPacketEventQueue();
        PacketEvent pe = null;
        int nameTrnId;

        try {
            while( retryCount-- > 0 ) {
                retryTimeout = UCAST_REQ_RETRY_TIMEOUT;
                ClientNameServicePacket.nameTrnId++;
                out.setNameTrnId();
                nameTrnId = out.getNameTrnId();

                Log.printPacketData( "name service packet", out );
                Log.printPacketDiagram( "name service packet diagram", out );
                Log.printHexDump( "name service packet hex dump",
                            out.getData(),
                            0,
                            out.getLength() );

                socket.send( out.packet );

                long start = System.currentTimeMillis();
                while( retryTimeout > 0 &&
                        ( pe = peq.getNextPacketEvent( (long)retryTimeout )) != null ) {
                    ClientNameServicePacket in =
                                        new ClientNameServicePacket( pe.getPacket() );
                    Log.printPacketData( "name service packet", in );
                    Log.printPacketDiagram( "name service packet diagram", in );
                    Log.printHexDump( "name service packet hex dump",
                            in.getData(),
                            0,
                            in.getLength() );
                    if( in.getNameTrnId() == nameTrnId &&
                            in.getPacketType() == ClientNameServicePacket.NODE_STATUS_RESPONSE ) {
                        socket.removePacketEventQueue( peq );
                        boolean isDataFromNodeStatus = addr.isDataFromNodeStatus;
                        NbtAddress[] addrs = in.getNodeNameArray( addr );
                        if( !isDataFromNodeStatus && !addr.isDataFromNodeStatus ) {
                            throw new UnknownNbtHostException();
                        }
                        return addrs;
                    }
                    retryTimeout -= System.currentTimeMillis() - start;
                }
            }
            socket.removePacketEventQueue( peq );
        } catch( IOException ioe ) {
        } catch( InterruptedException ie ) {
            socket = null;
            Log.printStackTrace( "jcifs.netbios.NbtNameServiceClient.getNodeStatus()", ie );
        }
        throw new UnknownNbtHostException( addr.getHostAddress() );
    }

    private NbtAddress doUnicast( ClientNameServicePacket nsp )
                                                throws IOException {

        NbtAddress addr = null;
        try {
            maxRedirectCount = MAX_REDIRECT_COUNT;
            addr = sendAndReceiveOne( nsp,
                                nameserver,
                                UNICAST );
        } catch( InterruptedException ie ) {
            socket = null;
            Log.printStackTrace( "jcifs.netbios.NbtNameServiceClient.doUnicast()", ie );
        }
        return addr;
    }
    private NbtAddress doBroadcast( ClientNameServicePacket nsp )
                                                throws IOException {

        NbtAddress addr = null;
        try {
            maxRedirectCount = MAX_REDIRECT_COUNT;
            addr = sendAndReceiveOne( nsp,
                                bCastAddress,
                                BROADCAST );
        } catch( InterruptedException ie ) {
            socket = null;
            Log.printStackTrace( "jcifs.netbios.NbtNameServiceClient.doBroadcast()", ie );
        }
        return addr;
    }

    private NbtAddress sendAndReceiveOne( ClientNameServicePacket out,
                                InetAddress host,
                                int transType )
                                throws IOException, InterruptedException {

        try {
            socket.start();
        } catch( SocketException se ) {
            Log.printStackTrace( "jcifs.netbios.NbtNameServiceClient.sendAndReceiveOne()", se );
            throw se;
        }
        PacketEventQueue peq = socket.createPacketEventQueue();
        PacketEvent pe = null;
        int retryCount;
        int retryTimeout;

        int nameTrnId = out.getNameTrnId();
        if( transType == UNICAST ) {
            if( nameserver == null ) {
                throw new UnknownHostException( "netbios name server must be set" );
            }
            out.setBroadcast( false );
            out.setAddress( nameserver );
            retryCount = UCAST_REQ_RETRY_COUNT;
        } else {
            if( bCastAddress == null ) {
                throw new UnknownHostException( "netbios broadcast address must be set" );
            }
            out.setBroadcast( true );
            out.setAddress( bCastAddress );
            retryCount = BCAST_REQ_RETRY_COUNT;
        }

        while( retryCount-- > 0 ) {
            if( transType == UNICAST ) {
                retryTimeout = UCAST_REQ_RETRY_TIMEOUT;
            } else {
                retryTimeout = BCAST_REQ_RETRY_TIMEOUT;
            }
            Log.printPacketData( "name service packet", out );
            Log.printPacketDiagram( "name service packet diagram", out );
            Log.printHexDump( "name service packet hex dump",
                            out.getData(),
                            0,
                            out.getLength() );
            socket.send( out.packet );

            long start = System.currentTimeMillis();
            while( retryTimeout > 0 &&
                    ( pe = peq.getNextPacketEvent( (long)retryTimeout )) != null ) {
                ClientNameServicePacket in =
                    new ClientNameServicePacket( pe.getPacket() );
                Log.printPacketData( "name service packet", in );
                Log.printPacketDiagram( "name service packet diagram", in );
                Log.printHexDump( "name service packet hex dump",
                            in.getData(),
                            0,
                            in.getLength() );
                if( in.getNameTrnId() == nameTrnId ) {
                    int packetType = in.getPacketType();
                    if( packetType == ClientNameServicePacket.POS_NAME_QUERY_RESPONSE ) {
                        socket.removePacketEventQueue( peq );
                        return in.getNbtAddress( out.name );
                    } if( transType == UNICAST ) {
                        switch( packetType ) {
                            case ClientNameServicePacket.NEG_NAME_QUERY_RESPONSE:
                                socket.removePacketEventQueue( peq );
                                return null;
                            case ClientNameServicePacket.WAIT_FOR_ACK_RESPONSE:
                                start = System.currentTimeMillis();
                                retryTimeout = in.getTTL();
                                if( retryTimeout > MAX_UCAST_REQ_RETRY_TIMEOUT ) {
                                    retryTimeout = MAX_UCAST_REQ_RETRY_TIMEOUT;
                                }
                                break;
                            case ClientNameServicePacket.REDIRECT_NAME_QUERY_RESPONSE:
                                if( maxRedirectCount-- < 1 ) {
                                    socket.removePacketEventQueue( peq );
                                    return null;
                                }
                                return sendAndReceiveOne( out, in.getInetAddress(), UNICAST );
                        }
                    }
                }
                retryTimeout -= System.currentTimeMillis() - start;
            }
        }
        socket.removePacketEventQueue( peq );
        return null;
    }

    private NbtAddress[] sendAndReceiveAll( ClientNameServicePacket out )
                                    throws IOException, InterruptedException {
        
        int retryCount = BCAST_REQ_RETRY_COUNT;
        int retryTimeout = BCAST_REQ_RETRY_TIMEOUT;
        int duration = BCAST_REQ_DURATION;
        int i = 0, n = 1;
        ClientNameServicePacket in;
        NbtAddress[] addrs = new NbtAddress[n];
        NbtAddress[] tmp;
        try {
            socket.start();
        } catch( SocketException se ) {
            Log.printStackTrace( "jcifs.netbios.NbtNameServiceClient.sendAndReceiveAll()", se );
            throw se;
        }
        PacketEventQueue peq = socket.createPacketEventQueue();
        PacketEvent pe = null;

        int nameTrnId = out.getNameTrnId();
        out.setBroadcast( true );

        while( i == 0 ) {
            if( retryCount-- <= 0 ) {
                socket.removePacketEventQueue( peq );
                return null;
            }
            Log.printPacketData( "name service packet", out );
            Log.printPacketDiagram( "name service packet diagram", out );
            Log.printHexDump( "name service packet hex dump",
                            out.getData(),
                            0,
                            out.getLength() );
            socket.send( out.packet );

            long start = System.currentTimeMillis();
            while(( pe = peq.getNextPacketEvent( (long)retryTimeout )) != null ) {
                in = new ClientNameServicePacket( pe.getPacket() );
                Log.printPacketData( "name service packet", in );
                Log.printPacketDiagram( "name service packet diagram", in );
                Log.printHexDump( "name service packet hex dump",
                            in.getData(),
                            0,
                            in.getLength() );
                if( in.getNameTrnId() == nameTrnId &&
                            in.getPacketType() == ClientNameServicePacket.POS_NAME_QUERY_RESPONSE ) {
                    addrs[i++] = in.getNbtAddress( out.name );
                    break;
                } else {
                    retryTimeout -= System.currentTimeMillis() - start;
                    if( retryTimeout < 1 ) {
                        pe = null;
                        break;
                    }
                }
            }
        }

        long start = System.currentTimeMillis();
        while(( pe = peq.getNextPacketEvent( duration )) != null ) {
            in = new ClientNameServicePacket( pe.getPacket() );
            Log.printPacketData( "name service packet", in );
            Log.printPacketDiagram( "name service packet diagram", in );
            Log.printHexDump( "name service packet hex dump",
                            in.getData(),
                            0,
                            in.getLength() );
            if( in.getNameTrnId() == nameTrnId &&
                        in.getPacketType() == ClientNameServicePacket.POS_NAME_QUERY_RESPONSE ) {
                if( i >= n ) {
                    tmp = new NbtAddress[ n *= 4 ];
                    System.arraycopy( addrs, 0, tmp, 0, i );
                    addrs = tmp;
                }
                addrs[i++] = in.getNbtAddress( out.name );
            } else {
                duration -= System.currentTimeMillis() - start;
                if( duration < 1 ) {
                    break;
                }
            }
        }
        socket.removePacketEventQueue( peq );

        if( i == n ) {
            tmp = addrs;
        } else {
            tmp = new NbtAddress[i];
            System.arraycopy( addrs, 0, tmp, 0, i );
        }
        return tmp;
    }
}

