/*
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Copyright 2000 by Michael B. Allen
 *
 */

package jcifs.util;

import java.net.DatagramSocket;
import java.net.DatagramPacket;
import java.net.SocketException;
import java.net.InetAddress;
import java.io.IOException;

/** 
 * <code>MultiDatagramSocket</code> employs a simple linked list queuing mechanism and an
 * internal thread dedicated to reading from the socket to allow multiple
 * threads to concurrently send and receive datagram packets on the same
 * port at rates independent from one another. There are three classes
 * that encapsulate this functionality: <code>MultiDatagramSocket</code>,
 * {@link PacketEventQueue}, and {@link PacketEvent}.
 * 
 * <p><img src='mds.gif' vspace='0' hspace='10' align='right' border='0'>
 *
 * MultiDatagramSocket is, at first glance, almost identical to
 * {@link java.net.DatagramSocket} but instead of calling <code>receive</code> you obtain a {@link jcifs.util.PacketEventQueue}
 * by calling {@link #createPacketEventQueue} and read in {@link
 * jcifs.util.PacketEvent}s at liesure by calling {@link PacketEventQueue#getNextPacketEvent}
 * on the <code>PacketEventQueue</code>. {@link jcifs.util.PacketEvent}
 * encapsulates the {@link java.net.DatagramPacket}, the reception time,
 * and is {@link java.lang.Cloneable}. The <code>PacketEventQueue.getNextPacketEvent()</code>
 * method will block until a packet is received if there are no
 * <code>PacketEvent</code>s in the queue.
 *
 * <p> This is useful when different tasks are servicing request
 * on the same port. You don't have to worry about tasks stealing
 * each other's packets. The netbios name service of jcifs
 * uses it to send and receive requests synchronously. Without
 * this, a failing request(which can take up to 15 seconds) would
 * prevent another request from being serviced until it finished.
 * This may all seem overkill but it is surprisingly easy to
 * implement. All three classes total a meager 250 lines of code. See <a
 * href='http://users.erols.com/mballen/MultiDatagramSocket.txt'>MultiDatagramSocket.txt</a>
 * for more information.
 * 
 * <p> <b>Important Note:</b> When the internal thread enqueues an incoming
 * packet it <code>clone</code>s it only once so that when another packet comes in the
 * data is not overwritten in the queues. But the data({@link java.net.DatagramPacket}, it's
 * buffer, and so on) are not <code>clone</code>d for entry into each queue. In other
 * words, with each incoming packet, the same {@link PacketEvent} object is queued
 * to all PacketEventQueues. So you CANNOT MODIFY the data of a PacketEvent
 * without cloning it first or other {@link PacketEventQueue}s will observe the
 * changes. This is one reason why <code>PacketEvent</code> is {@link java.lang.Cloneable}. This was done
 * for purposes of efficiency. I have found that in practice packets are
 * simply examined and not modified. So if you want to recycle a packet
 * for it's data just {@link PacketEvent#clone} it first.
 *
 * @version   0.1, 00/03/13
 * @author    Michael B. Allen
 * @see       PacketEventQueue
 * @see       PacketEvent
 * @see       java.net.DatagramSocket
 * @see       java.net.DatagramPacket
 * @see       java.lang.Cloneable
 * @since     jcifs-0.1
 */
 
public class MultiDatagramSocket implements Runnable {
    private static final int MAX_DATAGRAM_LENGTH = 576;
    private static final int DEFAULT_SO_TIMEOUT = 30000;
    private DatagramSocket socket;
    private int port, size;
    private InetAddress addr;
    private byte[] bufr;
    private DatagramPacket packet;
    private long time;
    private PacketEventQueue first, last, prev, tmp;
    private volatile boolean stop;
    private Thread t;
    private int soTimeout;

/**
 * Construct a new <code>MultiDatagramSocket</code>.
 */

    public MultiDatagramSocket() {
        this( 0, null );
    }

/**
 * Construct a new <code>MultiDatagramSocket</code> on the specified port.
 */

    public MultiDatagramSocket( int port ) {
        this( port, null );
    }

/**
 * Construct a new <code>MultiDatagramSocket</code> on the specified port
 * that binds to the specified interface.
 */

    public MultiDatagramSocket( int port, InetAddress addr ) {
        this.port = port;
        this.addr = addr;
        bufr = new byte[MAX_DATAGRAM_LENGTH];
        packet = new DatagramPacket( bufr, MAX_DATAGRAM_LENGTH );
        first = last = null;
        size = 0;
        stop = true;
        soTimeout = DEFAULT_SO_TIMEOUT;
    }

    public void run() {
        while( !stop ) {
            try {
                packet.setLength( MAX_DATAGRAM_LENGTH );
                socket.setSoTimeout( soTimeout );
                socket.receive( packet );
                time = System.currentTimeMillis();
            } catch( IOException ioe ) {
                socket.close();
                stop = true;
                Log.println( Log.WARNINGS,
                            "jcifs.util.MultiDatagramSocket warning",
                            " stopping datagram socket after " + ( soTimeout / 1000 ) +
                            " seconds of no activity" );
                break;
            }
            postPacketEvent( (PacketEvent)( new PacketEvent( packet, time ).clone() ));
        }
    }

//
// POSSIBLE ERROR CONDITON: What will happen if the socket times out while a thread is still processing items in its PacketEventQueue?
//

/**
 * Start this <code>MultiDatagramSocket</code>. After 30 seconds of no activity the socket
 * will timeout and close. Should another packet arrive the socket will
 * restart and service the request as normal.
 */

    public void start() throws SocketException {
        if( !stop ) {
            return;
        }
        socket = new DatagramSocket( port, addr );
        stop = false;
        t = new Thread( this );
//        t.setPriority( Thread.NORM_PRIORITY + 1 );
        Log.println( Log.WARNINGS,
                            "jcifs.util.MultiDatagramSocket warning",
                            " starting datagram socket on port " +
                            socket.getLocalPort() );
        t.start();
    }

/**
 * Stop the <code>MultiDatagramSocket</code>. You can start and stop <code>MultiDatagramSocket</code> any number
 * of times without worrying about reinstantiating.
 */

    public void stop() {
        if( stop ) {
            return;
        }
        Log.println( Log.WARNINGS,
                    "jcifs.util.MultiDatagramSocket warning",
                    " stopping datagram socket" );
        stop = true;
        socket.close();
        interruptPacketEventQueues();
        try {
            t.join( 2000 );
        } catch( InterruptedException ie ) {
            Log.printStackTrace( "jcifs.util.MultiDatagramSocket exception", ie );
        }
    }

/**
 * Send a datagram.
 *
 * @param  p the <code>DatagramPacket</code> to send
 * @throws java.io.IOException if the send operation fails
 */

    public void send( DatagramPacket p ) throws IOException {
        socket.send( p );
    }

/**
 * Retrieves the local hosts interface this socket is running bound to.
 */

    public InetAddress getLocalAddress() {
        return socket.getLocalAddress();
    }

/**
 * Returns the local port this socket is set to run on.
 */ 

    public int getLocalPort() {
        return socket.getLocalPort();
    }

/**
 * Create a new {@link PacketEventQueue}. As soon as a thread creates a new <code>PacketEventQueue</code> it is
 * currently queueing packets. So if the thread chooses not to call {@link
 * PacketEventQueue#getNextPacketEvent} for a while it may get
 * a back log of packets. This is probably not what you want. Obtain the
 * {@link PacketEventQueue} just before the calling thread is going to try
 * to retrive a packet.
 */

    public synchronized PacketEventQueue createPacketEventQueue() {
        PacketEventQueue peq = new PacketEventQueue();
        size++;

        if( first == null ) {
            first = last = peq;
            return peq;
        }
        peq.next = first;
        first = peq;
        return peq;
    }

/**
 * Discard the {@link PacketEventQueue}(ie stop listening). It is always
 * very importantant to release a <code>PacketEventQueue</code> or the
 * socket will maintain it's reference and queue packet events to it.
 * Probably get a VM memory error eventually.
 */

    public synchronized void removePacketEventQueue( PacketEventQueue peq ) {
        if( first == null ) {
            return;
        }
        tmp = first;
        while( tmp != null ) {
            if( tmp.equals( peq )) {
                if( tmp.equals( first )) {
                    first = tmp.next;
                } else {
                    prev.next = tmp.next;
                    if( tmp.equals( last )) {
                        last = prev;
                    }
                }
                size--;
                return;
            }
            prev = tmp;
            tmp = tmp.next;
        }
    }
    synchronized void postPacketEvent( PacketEvent pe ) {
        if( first == null ) {
            return;
        }
        tmp = first;
        do {
            tmp.add( pe );
            tmp = tmp.next;
        } while( tmp != null );
    }

/**
 * Interrupt any threads waiting for {@link PacketEvent}s on their {@link PacketEventQueue}s.
 */

    public synchronized void interruptPacketEventQueues() {
        if( first == null ) {
            return;
        }
        tmp = first;
        do {
            tmp.interrupt();
            tmp = tmp.next;
        } while( tmp != null );
    }

/**
 * Checks if there are any open {@link PacketEventQueue}s.
 */

    public boolean isEmpty() {
        return ( first == null );
    }

/**
 * The number of open {@link jcifs.util.PacketEventQueue}s.
 */

    public int getSize() {
        return size;
    }
}

