/*
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Copyright 2000 by Michael B. Allen
 *
 */

package jcifs.util;

/**
 * This class is a container for {@link PacketEvent}s and is obtained from
 * {@link MultiDatagramSocket}.
 * 
 * @version   0.1, 00/03/13
 * @author    Michael B. Allen
 * @see MultiDatagramSocket
 * @see PacketEvent
 * @since     jcifs-0.1
 */

public class PacketEventQueue {
    private PacketEvent first;
    private PacketEvent last;
    private int size;
    private Thread currentThread = null;

    PacketEventQueue next;

    PacketEventQueue() {
        first = last = null;
        size = 0;
    }

    synchronized void add( PacketEvent pe ) {
        if( first == null ) {
            first = last = pe;
        } else {
            last = last.next = pe;
        }
        size++;
        notify();
    }

/**
 * Retreive or wait for the next {@link PacketEvent}. This will block
 * if there are no packets in the queue.
 *
 * @throws java.lang.InterruptedException if the <code>MultiDatagramSocket</code> is closed or {@link MultiDatagramSocket#interruptPacketEventQueues} is called
 */

    public synchronized PacketEvent getNextPacketEvent()
                                            throws InterruptedException {
        return getNextPacketEvent( 0 );
    }

/**
 * Retreive or wait for the next {@link PacketEvent} for at most
 * <code>timeout</code> milliseconds.
 *
 * @param timeout the amount of time to wait blocked
 * @throws InterruptedException if the socket is closed or the <code>timeout</code> period is reached
 */

    public synchronized PacketEvent getNextPacketEvent( long timeout )
                                            throws InterruptedException {
        while( first == null ) {
            try {
                currentThread = Thread.currentThread();
                wait( timeout );
                currentThread = null;
            } catch( InterruptedException ie ) {
                currentThread = null;
                throw ie;
            }
            if( first == null ) {
                return null;
            }
        }
        PacketEvent pe = first;
        first = first.next;
        size--;
        return pe;
    }
    synchronized void interrupt() {
        if( currentThread != null ) {
            currentThread.interrupt();
        }
    }

/**
 * Determine if the queue is empty.
 */

    public boolean isEmpty() {
        return first == null;
    }

/**
 * Return the number of packets in the queue
 */

    public int getSize() {
        return size;
    }

/**
 * Clear all {@link PacketEvent}s in the <code>PacketEventQueue</code>
 */

    public void clear() {
        first = last = null;
        size = 0;
    }
}

